"""Module containing logic for describe-get-system proof of concept."""


class Dgs:
    """Describe-Get-System class"""
    def __init__(self, *args, **kwargs):    # noqa
        NotImplementedError('TODO: Need to implement Dgs class')
